#include <system.h>
#include "diskio.h"
#include "serial.h"
#include "tff.h"

/* ATA command */
#define CMD_RESET		0x08	/* DEVICE RESET */
#define CMD_READ		0x20	/* READ SECTOR(S) */
#define CMD_WRITE		0x30	/* WRITE SECTOR(S) */
#define CMD_IDENTIFY	0xEC	/* DEVICE IDENTIFY */
#define CMD_SETFEATURES	0xEF	/* SET FEATURES */

/* ATA register bit definitions */
#define	LBA				0xE0
#define	BSY				0x80
#define	DRDY			0x40
#define	DF				0x20
#define	DRQ				0x08
#define	ERR				0x01
#define	SRST			0x04
#define	nIEN			0x02


/* Bit definitions for Control Port */
#define	CTL_READ		0x20
#define	CTL_WRITE		0x40
#define	CTL_RESET		0x80
#define	REG_DATA		0xF0
#define	REG_ERROR		0xF1
#define	REG_FEATURES	0xF1
#define	REG_COUNT		0xF2
#define	REG_SECTOR		0xF3
#define	REG_CYLL		0xF4
#define	REG_CYLH		0xF5
#define	REG_DEV			0xF6
#define	REG_COMMAND		0xF7
#define	REG_STATUS		0xF7
#define	REG_DEVCTRL		0xEE
#define	REG_ALTSTAT		0xEE

#define READ 0
#define WRITE 1
#define ENABLE 1
#define DISABLE 0

#define CS_PORT porta
#define CS_DDR trisa
#define FLG_PORT portb
#define FLG_DDR trisb
#define EN_PORT portc
#define EN_DDR trisc
#define DATA_PORT portd
#define DATA_DDR trisd
#define ADR_PORT porte
#define ADR_DDR trise

#define CFCE1 1
#define CFCE2 2
#define CFINT 2
#define IORDY 3
#define IOE 0
#define BIT_IOW 1
#define BIT_IOR 2

inline void SetCS(char adr)
{
	if(adr & 0x08)
		set_bit(CS_PORT,CFCE1);
	else
		clear_bit(CS_PORT,CFCE1);
	if(adr & 0x10)
		set_bit(CS_PORT,CFCE2);
	else
		clear_bit(CS_PORT,CFCE2);
}

static volatile
DSTATUS Stat = STA_NOINIT&(~STA_NODISK);	/* Disk status */

static volatile
BYTE Timer;	/* 100Hz decrement timer */
	
/*******************************
**  D[15..8]̃b`
**  Fo͂̕(READ/WRITE)
**  ߂lFȂ
********************************/
void latch_high_byte(char dir)
{
	if(dir == READ)
	{
		if(test_bit(EN_PORT,IOE))
			clear_bit(EN_PORT,IOE);
		ADR_PORT = 0x00;
		set_bit(EN_PORT,IOE);
		clear_bit(EN_PORT,IOE);
	}
	else if(dir == WRITE)
	{
		if(test_bit(EN_PORT,IOE))
			clear_bit(EN_PORT,IOE);
		ADR_PORT = 0x01;
		set_bit(EN_PORT,IOE);
		clear_bit(EN_PORT,IOE);
	}
}

/*******************************
**  D[15..8]̃Zbg
**  Fo͂̕(READ/WRITE),o͋(ENABLE/DISABLE)
**  ߂lFȂ
********************************/
void set_high_byte(char dir,char en)
{
	if(dir == READ)
	{
		if(en == ENABLE)
		{
			ADR_PORT = 0x02;
			set_bit(EN_PORT, IOE);
		}
		else if(en == DISABLE)
		{
			clear_bit(EN_PORT, IOE);
		}
	}
	else if(dir==WRITE)
	{
		if(en == ENABLE)
		{
			clear_bit(EN_PORT, BIT_IOW);
		}
		else if(en == DISABLE)
		{
			set_bit(EN_PORT, BIT_IOW);
		}
	}
}



/*******************************
**  ATAWX^ǂݏo
**  FAhX
**  ߂lFWX^̓e
********************************/
BYTE read_ata(BYTE adr)
{

	BYTE rd_dat;
	DATA_DDR = 0xFF;
	//CS_PORT &= ~(0x06)|(adr>>2)&0x06;	//CS[1..0] assert
	SetCS(adr);
	ADR_PORT = (0x07 & adr);		//ADR[2..0] assert
	clear_bit(EN_PORT, BIT_IOR);	//IOR assert
	//delay_10us(1);
	rd_dat = DATA_PORT;		//D[7..0][h
	//delay_10us(1);
	set_bit(EN_PORT, BIT_IOR);	//IOR negate
	//CS_PORT |= 0x06;
	SetCS(0xff);
	return rd_dat;
}

/*******************************
**  ATAWX^
**  FAhX, ݃f[^
**  ߂lFȂ
********************************/
void write_ata(BYTE adr, BYTE dat)
{
	DATA_DDR = 0x00;
	//CS_PORT &= ~(0x06)|(adr>>2)&0x06;	//CS[1..0] assert
	SetCS(adr);

	ADR_PORT = (0x07 & adr);		//ADR[2..0] assert
	clear_bit(EN_PORT, BIT_IOW);	//IOW assert
	//delay_10us(1);
	DATA_PORT = dat;		//D[7..0]Cg
	//delay_10us(1);
	set_bit(EN_PORT, BIT_IOW);	//IOW negate
	//CS_PORT |=0x06;
	SetCS(0xff);
	DATA_DDR = 0xFF;
	//delay_10us(1);
}

/*---------------------*/
/* Wait for Data Ready */


static
BOOL wait_data ()
{
	BYTE s;


	Timer = 100;	/* Time out = 1 sec */
	do {
		if (!Timer) return FALSE;	/* Abort when timeout occured */
		s = read_ata(REG_STATUS);	/* Get status */
	} while (s & BSY);				/* Repeat until BSY goes low */

	if ((s & ERR)) return FALSE;	/* Abort when no data given */

	read_ata(REG_ALTSTAT);
	return TRUE;
}



/*---------------------------*/
/* Read a part of data block */

static
void read_part (
	BYTE *buff, 	/* Data buffer to store read data */
	BYTE ofs,		/* Offset of the part of data in unit of word */
	BYTE count		/* Number of word to pick up */
)
{
	BYTE c = 0, dl, dh;



	//CS_PORT &= (REG_DATA>>2)|0xF9;	//CS[1..0] assert
	SetCS(REG_DATA);
	ADR_PORT = (0x07 & REG_DATA);		//ADR[2..0] assert
	do {
	    clear_bit(EN_PORT,BIT_IOR);	//IOR assert
		//delay_10us(1);
	    dl = DATA_PORT;		//D[7..0][h`
	    set_bit(EN_PORT,BIT_IOR);	//IOR negate
		//delay_10us(1);
	    clear_bit(EN_PORT,BIT_IOR);	//IOR assert
		//delay_10us(1);
	    dh = DATA_PORT;//D[15..8][h
	    set_bit(EN_PORT,BIT_IOR);	//IOR negate
		//delay_10us(1);
		if (count && (c >= ofs)) {	/* Pick up a part of block */
			*buff++ = dl;
			*buff++ = dh;
			count--;
		}
	} while (++c);
	//CS_PORT |=0x06;
	SetCS(0xff);
	read_ata(REG_ALTSTAT);
	read_ata(REG_STATUS);
}


/*******************************
**  ATAf[^
**  FȂ
**  ߂lFXe[^X
********************************/
DSTATUS disk_initialize(void)
{
  Stat |= STA_NOINIT;
  
  
  // |[g̐ݒ

  //CS_PORT |= (1<<CFCE1)|(1<<CFCE2);
  //CS_DDR &= ~(1<<CFCE1) & ~(1<<CFCE2);
  SetCS(0xff);
  FLG_DDR |= (1<<IORDY)|(1<<CFINT);
  EN_PORT |= (1<<BIT_IOR)|(1<<BIT_IOW);
  EN_DDR &= ~(1<<BIT_IOR)&~(1<<BIT_IOW)&~(1<<IOE);
  DATA_DDR = 0xFF;
  ADR_PORT &= 0xF8;
  ADR_DDR &= 0xF8;
  set_high_byte(READ,DISABLE);
  set_high_byte(WRITE,DISABLE);
  
  write_ata(REG_DEV, 0);

  Timer = 200;
  do{
    if(!Timer)return Stat;
  }while(read_ata(REG_STATUS) & (BSY | DRQ));
  


  write_ata(REG_DEVCTRL, SRST | nIEN);	/* \tgZbg */
  for(Timer = 2; Timer; );
  write_ata(REG_DEVCTRL, nIEN);
  for(Timer = 2; Timer; );
  Timer = 100;
  do{
	if(!Timer)return Stat;
  }while((read_ata(REG_STATUS) & (DRDY|BSY)) != DRDY);

  
  write_ata(REG_FEATURES, 0x01);	/* 8bit[hANZX */
  write_ata(REG_COMMAND, CMD_SETFEATURES);
  Timer = 100;
  do{
    if(!Timer)return Stat;
  }while(read_ata(REG_STATUS) & BSY);
  
  Stat &= ~STA_NOINIT;
  
  return Stat;
}

DSTATUS disk_status(void)
{
	return Stat;
}

/*----------------*/
/* Read Sector(s) */

DRESULT disk_read (
	DiskBuff rbuff,		/* Data buffer to store read data */
	DWORD sector,	/* Sector number (LBA) */
	BYTE count		/* Sector count (1..255) */
)
{
	BYTE c, iord_l, iord_h,i;
	//BYTE *buff_b=buff->lbuf;
    BYTE *buff;

	if (Stat & STA_NOINIT) return RES_NOTRDY;
	if (count!=1) return RES_PARERR;  //only support count==1
	/* Issue Read Setor(s) command */
	write_ata(REG_COUNT, count);
	write_ata(REG_SECTOR, (BYTE)sector);
	write_ata(REG_CYLL, (BYTE)(sector >> 8));
	write_ata(REG_CYLH, (BYTE)(sector >> 16));
	write_ata(REG_DEV, ((BYTE)(sector >> 24) & 0x0F) | LBA);
	write_ata(REG_COMMAND, CMD_READ);

	do {
		if (!wait_data()){
			SetCS(0xff);
			return RES_ERROR;	/* Wait data ready */
		}
		//CS_PORT = (REG_DATA>>2)&0x06;	//CS[1..0] assert
		SetCS(REG_DATA);
		ADR_PORT = (0x07 & REG_DATA);		//ADR[2..0] assert
        //Read Low
        buff=rbuff.lbuff;
        c = 0;
		do {
			clear_bit(EN_PORT,BIT_IOR);	//IOR assert
			//delay_10us(1);
			*buff++ = DATA_PORT;		//D[7..0][h
			set_bit(EN_PORT, BIT_IOR);	//IOR negate
			//delay_10us(1);
        } while (--c);
        //Read high
        buff=rbuff.hbuff;
        c = 0;
        do {
			clear_bit(EN_PORT,BIT_IOR);	//IOR assert
			//delay_10us(1);
			*buff++ = DATA_PORT;//D[15..8][h
			set_bit(EN_PORT, BIT_IOR);	//IOR negate
			//delay_10us(1);
		} while (--c);
	} while (--count);

	read_ata(REG_ALTSTAT);
	read_ata(REG_STATUS);
	SetCS(0xff);
	return RES_OK;
}

/*--------------------------*/
/* Miscellaneous Functions  */

DRESULT disk_ioctl (
	BYTE ctrl,		/* Control code */
	BYTE *buff		/* Buffer to send/receive data block */
)
{
	BYTE n, w, ofs, dl, dh, *ptr = buff;


	if (Stat & STA_NOINIT) return RES_NOTRDY;

	switch (ctrl) {
		case GET_SECTORS :	/* Get number of sectors on the disk (unsigned long) */
			ofs = 60; w = 2; n = 0;
			break;

		case ATA_GET_REV :	/* Get firmware revision (8 chars) */
			ofs = 23; w = 4; n = 4;
			break;

		case ATA_GET_MODEL :	/* Get model name (40 chars) */
			ofs = 27; w = 20; n = 20;
			break;

		case ATA_GET_SN :	/* Get serial number (20 chars) */
			ofs = 10; w = 10; n = 10;
			break;

		default:
			return RES_PARERR;
	}

	write_ata(REG_COMMAND, CMD_IDENTIFY);
	if (!wait_data()) return RES_ERROR;
	read_part(ptr, ofs, w);
	while (n--) {
		dl = *ptr; dh = *(ptr+1);
		*ptr++ = dh; *ptr++ = dl; 
	}

	return RES_OK;
}

void disk_timerproc ()
{
	static BYTE pv;
	BYTE n;


	n = Timer;					/* 100Hz decrement timer */
	if (n) Timer = --n;
}
